
import sys, os

[ BT_None, BT_Unconditional, BT_Conditional, BT_Serialising ] = "NUCS"
BT_LOOKUP = {
        "None" : BT_None,
        "Unconditional" : BT_Unconditional,
        "Conditional" : BT_Conditional,
        "Serialising" : BT_Serialising,
        }
[ BC_Invalid, BC_Stack, BC_Heap ] = "ish"
[ AP_NOT_ITERATIVE, AP_PSEUDO_STATIC, AP_CONSTANT_STEP,
        AP_VARIABLE_STEP, AP_RANDOM_ACCESS ] = "npcvr"
IDENTIFY_CLASS = {
        "nit": AP_NOT_ITERATIVE,
        "pst": AP_PSEUDO_STATIC,
        "cst": AP_CONSTANT_STEP,
        "vst": AP_VARIABLE_STEP,
        "rnd": AP_RANDOM_ACCESS,
}

class CFG_Vertex:
    def __init__(self, id):
        self.id = id
        self.pc = None
        self.entrances_pass_one = []
        self.exits_pass_one = []
        self.present_pass_one = []
        self.entrances = None
        self.exits = None
        self.present = None
        self.udg_printed = False
        self.identify_loops = False
        self.is_exit = False
        self.depth = 0

    def Import_Data(self, fields, table):
        key = fields[ 0 ]
        if ( key == "entrance" ):
            table[ key ] = self
            return
        elif ( key == "exit" ):
            table[ key ] = self
            self.is_exit = True
            return

        value = int(fields[ 1 ], 16)
        if ( key == "pc" ):
            self.pc = value
        elif ( key == "ie" ):
            self.entrances_pass_one.append(value)
        elif ( key == "oe" ):
            self.exits_pass_one.append(value)
        elif ( key == "present" ):
            self.present_pass_one.append(value)

    def Pass_Two(self, table):
        self.entrances = []
        self.exits = []
        self.present = []
        for i in self.entrances_pass_one:
            self.entrances.append(table[ i ])
        for i in self.exits_pass_one:
            self.exits.append(table[ i ])
        for i in self.present_pass_one:
            self.present.append(table[ i ])
            self.present[ -1 ].present.append(self)
        self.entrances_pass_one = None
        self.exits_pass_one = None
        self.present_pass_one = None

    def Clear_UDG_Flag(self):
        self.udg_printed = False

    def Export_UDG(self, udg_out, edge_label_fn, 
                    vertex_label_fn=None, root=True):
        if self.udg_printed:
            if root:
                return False
            else:
                udg_out.write(r'r("L%x")' % self.id)
                return True

        self.udg_printed = True
        vertex_label = ""
        if vertex_label_fn != None:
            vertex_label = vertex_label_fn(self)

        udg_out.write('l("L%x",n("C%x",[a("OBJECT","%s")],[' % (self.id, 
                self.id, vertex_label))

        first = True
        for e in self.exits:
            if not first:
                udg_out.write(',')

            first = False
            udg_out.write('e("E%x",[a("OBJECT","")],'
                'l("Z%x",n("Y%x",[a("OBJECT","%s")],['
                'e("X%x",[a("OBJECT","")],' % (e.id, e.id, e.id,
                        edge_label_fn(e), e.id))
            e.end.Export_UDG(udg_out, edge_label_fn, vertex_label_fn, False)
            udg_out.write(')])))')

        udg_out.write(']))')
        return True


class CFG_Edge:
    def __init__(self, id):
        self.id = id
        self.frequency = self.start = self.end = None
        self.path = []
        self.branch_type = BT_None
        self.superblock_next = None
        self.transition_limit = 10000
        self.superblock = None
        self.base_cost = 0
        self.cost = 0
        self.max_frequency = None
        self.pc = None
        self.end_pc = None
        self.loop_body = None
        self.loop_header = None
        self.is_dirty = False
        self.mem_ops = set()
        self.creator_pass_one = []
        self.user_pass_one = []
        self.creator = None
        self.user = None

    def Pass_Two(self, table):
        self.creator = []
        self.user = []
        for i in self.creator_pass_one:
            self.creator.append(table[ i ])
            self.creator[ -1 ].creators.append(self)
        for (i, pc_list) in self.user_pass_one:
            self.user.append((table[ i ], pc_list))
            table[ i ].users.append((self, pc_list))
        self.creator_pass_one = None
        self.user_pass_one = None

    def Import_Data(self, fields, table):
        key = fields[ 0 ]
        try:
            value = int(fields[ 1 ], 16)
        except:
            value = fields[ 1 ]

        if ( key == "f" ):
            self.frequency = value
            self.max_frequency = value
        elif ( key == "iv" ):
            self.start = table[ value ]
        elif ( key == "ov" ):
            self.end = table[ value ]
        elif ( key == "x" ):
            if ( self.pc == None ):
                self.pc = value
            self.path.append(value)
            self.end_pc = value + 4
            if (fields[ -1 ] == "mem"):
                self.mem_ops.add(value)
        elif ( key == "cost" ):
            self.base_cost = self.cost = value
        elif ( key == "bt" ):
            self.branch_type = BT_LOOKUP[ value ]
        elif ( key == "creator" ):
            self.creator_pass_one.append(value)
        elif ( key == "user" ):
            self.user_pass_one.append((value, 
                    [ int(x, 16) for x in fields[ 2: ] ]))

class CFG_Base_Creator:
    def __init__(self, id):
        self.id = id
        self.type = BC_Invalid
        self.max = 0
        self.size = 0
        self.users = []
        self.creators = []
        self.present = []
        self.sample = None
        self.ap_data = []

    def Import_Data(self, fields, table):
        key = fields[ 0 ]
        try:
            value = int(fields[ 1 ], 16)
        except:
            value = fields[ 1 ]

        if ( key == "type" ):
            if ( value == "valid" ):
                self.type = BC_Heap
            elif ( value == "stack" ):
                self.type = BC_Stack
            elif ( value == "invalid" ):
                self.type = BC_Invalid
            else:
                assert False
        elif ( key == "max" ):
            self.max = value
        elif ( key == "size" ):
            self.size = value
        elif ( key == "sample" ):
            self.sample = value
        elif ( key == "class" ):
            self.ap_data.append({
                    "pc" : value,
                    "ap_class" : IDENTIFY_CLASS[ fields[ 2 ] ],
                    "ap_constant_step" : int(fields[ 3 ]),
                    "sample" : int(fields[ 4 ], 16),
                    "counter" : int(fields[ 5 ], 16),
                })

def Find_V_E(entrance):
    assert isinstance(entrance, CFG_Vertex)

    out_vertexes = []
    out_edges = []
    seen_before = set()
    todo = [ entrance ]

    while ( len(todo) != 0 ):
        element = todo.pop()
        if ( element.id in seen_before ):
            continue
        
        seen_before.add(element.id)

        if ( isinstance(element, CFG_Vertex) ):
            out_vertexes.append(element)
            for edges in (element.entrances, element.exits):
                for edge in edges:
                    todo.append(edge)

        elif ( isinstance(element, CFG_Edge) ):
            out_edges.append(element)
            if ( element.start != None ):
                todo.append(element.start)

            if ( element.end != None ):
                todo.append(element.end)

        else:
            assert False, "CFG element is not Edge or Vertex"

    return (out_vertexes, out_edges)
    
def Find_Base_Creators(entrance):
    assert isinstance(entrance, CFG_Vertex)

    out_bcs = []
    seen_before = set()
    todo = [ entrance ]

    while ( len(todo) != 0 ):
        element = todo.pop()
        if ( element.id in seen_before ):
            continue
        
        seen_before.add(element.id)

        if ( isinstance(element, CFG_Vertex) ):
            for edges in (element.entrances, element.exits):
                for edge in edges:
                    todo.append(edge)

        elif ( isinstance(element, CFG_Edge) ):
            if ( element.start != None ):
                todo.append(element.start)

            if ( element.end != None ):
                todo.append(element.end)

            for bc in element.creator:
                if ( not ( bc.id in seen_before )):
                    seen_before.add(bc.id)
                    out_bcs.append(bc)

            for (bc, pc_list) in element.user:
                if ( not ( bc.id in seen_before )):
                    seen_before.add(bc.id)
                    out_bcs.append(bc)

        else:
            assert False, "CFG element is not Edge or Vertex"

    return out_bcs
    

def Read_CFG(fname, vertex_class=CFG_Vertex, edge_class=CFG_Edge,
            base_creator_class=CFG_Base_Creator):
    assert issubclass(vertex_class, CFG_Vertex)
    assert issubclass(edge_class, CFG_Edge)
    assert issubclass(base_creator_class, CFG_Base_Creator)
    vertex = edge = bc = None
    type_code = '?'
    vlist = []
    elist = []
    table = dict()
    for line in file(fname):
        fields = line.split()

        if (( len(line) == 0 )
        or ( line[ 1 ] != ' ' )):
            type_code = '?'

        elif ( line[ 0 ] == 'v' ):
            type_code = line[ 0 ]
            id = int(fields[ 1 ], 16)
            vertex = vertex_class(id)
            table[ id ] = vertex
            vlist.append(vertex)

        elif ( line[ 0 ] == 'e' ):
            type_code = line[ 0 ]
            id = int(fields[ 1 ], 16)
            edge = edge_class(id)
            table[ id ] = edge
            elist.append(edge)

        elif ( line[ 0 ] == 'b' ):
            type_code = line[ 0 ]
            id = int(fields[ 1 ], 16)
            bc = base_creator_class(id)
            table[ id ] = bc

        elif ( line[ :2 ] == "  " ):
            if ( type_code == 'v' ):
                vertex.Import_Data(fields, table)
            elif ( type_code == 'e' ):
                edge.Import_Data(fields, table)
            elif ( type_code == 'b' ):
                bc.Import_Data(fields, table)

        else:
            type_code = '?'

    for vertex in vlist:
        vertex.Pass_Two(table)
    for edge in elist:
        edge.Pass_Two(table)
    
    return table[ "entrance" ]

def Identify_Loops(entrance):
    assert isinstance(entrance, CFG_Vertex)
    loop_list = []
    edge_history = []
    loop_set = set()
 
    def Vertex(v1, depth):
        if len(v1.entrances) > 1:
            # Join point
            assert len(v1.exits) == 1
            loop_header = None
            loop_body = None
            for i in xrange(len(edge_history) - 1):
                e = edge_history[ i ]
                if v1 == e.end:
                    loop_header = e
                    loop_body = edge_history[ i + 1 : ]
                    break

            if loop_header != None:
                edge1 = v1.exits[ 0 ]
                edge1.loop_header = loop_header
                assert edge1 == loop_body[ 0 ]
                loop_header.loop_body = loop_body
                if not (loop_header.id in loop_set):
                    loop_list.append(loop_header)
                    loop_set.add(loop_header.id)
        else:
            # Not a join point
            pass

        if v1.identify_loops:
            if depth >= v1.depth:
                # reached via a longer/same distance path
                return
            else:
                # reached via a shorter path, reprocess please
                pass
        else:
            v1.identify_loops = True 

        v1.depth = depth
        depth += 1

        for e1 in v1.exits:
            if e1.loop_header == None:
                edge_history.append(e1)
                Vertex(e1.end, depth)
                e2 = edge_history.pop()
                assert e2 == e1


    Vertex(entrance, 0)
    return loop_list



