
from ipet import Common_Solver, Common_Setup, Name
from readcfg import CFG_Vertex, CFG_Edge
from readcfg import BT_Serialising, BT_Unconditional, BT_Conditional
import ilp, readcfg

class Superblock_Edge:
    def __init__(self, supername, source, edge, number, impossible_flag=False):
        self.source = source
        self.edge = edge
        self.cost = 0
        self.sinks = []
        self.name = "%sI%u" % (supername, number)
        self.frequency = None
        self.impossible_flag = impossible_flag
        if source != None:
            source.sinks.append(self)

    def Get_Next(self, ignore_me):
        if (self.edge != None) and (self.edge.pc != None) and not ignore_me:
            return [(self, self.cost, [self])]

        out = []
        for s in self.sinks:
            for (sb_edge, cost, p) in s.Get_Next(False):
                if self.cost != None:
                    cost += self.cost
                out.append((sb_edge, cost, 
                        [self] + p))

        return out

class Superblock:
    def __init__(self, start_edge):
        self.start_edge = start_edge
        self.Clear()

    def Clear(self):
        self.internal_dict = dict()
        self.number = 0
        self.name = "S%s" % Name(self.start_edge)
        self.main_exit = self.root = Superblock_Edge(
                    self.name, None, None, self.number)
        self.side_exits = []
        self.followers = []
        self.side_exit_ptr = 0
        self.number += 1

    def Add_Side_Exit(self, e_path, e_exit_list):
        # Add edge e_path to path, and members of e_exit_list as a side exit.
        # All of these side exits will be represented by a single entry
        # in self.side_exits on the grounds that all represent destinations
        # from the same multibranch
        never_taken = (len(e_exit_list) == 0)
        fork = Superblock_Edge(
                    self.name, self.main_exit, e_path, self.number)
        self.main_exit = Superblock_Edge(
                    self.name, fork, None, self.number + 1)
        fork = Superblock_Edge(
                    self.name, fork, None, self.number + 2, never_taken)
        self.side_exits.append(fork)
        self.number += 3
        for e_exit in e_exit_list:
            Superblock_Edge(self.name, fork, e_exit, self.number)
            self.number += 1

    def Add_Followers(self, e_exit_list):
        x = None
        for e_exit in e_exit_list:
            x = Superblock_Edge(self.name, self.main_exit, 
                        e_exit, self.number)
            self.followers.append(x)
            self.number += 1

        if x != None:
            self.main_exit = self.followers.pop()

    def Add_No_Exit(self, e):
        self.main_exit = Superblock_Edge(
                    self.name, self.main_exit, e, self.number)
        self.number += 1

    def Add_Side_Exit_Cost(self, cost):
        if ( self.side_exit_ptr < len(self.side_exits) ):
            self.side_exits[ self.side_exit_ptr ].cost = cost
            self.side_exit_ptr += 1
        else:
            self.main_exit.cost = max(self.main_exit.cost, cost)
            for f in self.followers:
                f.cost = self.main_exit.cost

    def Add_Result(self, key, value):
        assert self.internal_dict.has_key(key)
        sb_edge = self.internal_dict[ key ]
        sb_edge.frequency = value

    def Get_Frequency(self, key):
        return self.internal_dict[ key ].frequency

    def Draw(self, debug_out):
        def RecursiveDraw(root, left):
            if root.edge == None:
                match = " - None"
            else:
                match = " - %s (%x, %d)" % (Name(root.edge), 
                            root.edge.pc, len(root.edge.path))

            if len(root.sinks) == 0:
                match = " - exit" + match

            if root.impossible_flag or (root.cost == None):
                assert root.frequency in (0, None)
                match = " - impossible" + match
            else:
                assert type(root.cost) == int
                if type(root.frequency) == int:
                    match = '*%d%s' % (root.frequency, match)
                else:
                    match = '*??%s' % (match)

                match = ' - %d%s' % (root.cost, match)

            debug_out.write('%s+- %s%s\n' % (left, root.name, match))

            remaining = len(root.sinks)
            for sink in root.sinks:
                if remaining == 1:
                    RecursiveDraw(sink, "%s  " % left)
                else:
                    RecursiveDraw(sink, "%s| " % left)
                remaining -= 1

        debug_out.write('superblock %x - ?*%d\n' % (self.start_edge.pc, 
                                self.start_edge.frequency))
        RecursiveDraw(self.root, "")

    def Collect_Edges(self, udg_edge_list_dict):
        def RecursiveCollect(root):
            if root.edge != None:
                out = udg_edge_list_dict.get(root.edge.id, [])
                out.append(root)
                udg_edge_list_dict[ root.edge.id ] = out
            
            for sink in root.sinks:
                RecursiveCollect(sink)

        RecursiveCollect(self.root)

    def Add_To_Model(self, model, wcet, superblock_dict):
        #print 'ATM'

        def Recursive(root):
            model.Add_Var(root.name, ilp.VAR_INTEGER)
            superblock_dict[ root.name ] = self
            self.internal_dict[ root.name ] = root

            # all nodes - tie to f(x) values
            if root.edge != None:
                root.edge.contexts.append(root.name)
                #print hex(root.edge.pc)

            targets = []

            for sink in root.sinks:
                Recursive(sink)
                targets.append(sink.name)
        
            if len(targets) != 0:
                # non-leaf node - express tree structure
                model.Add_Term(root.name, "=", '+'.join(targets))

            if root.impossible_flag or (root.cost == None):
                # This should never be executed
                #print 'none', root.name
                model.Add_Term(root.name, '=', '0')
            else:
                #print root.cost, root.name
                assert type(root.cost) == int
                if root.cost > 0:
                    wcet.append('%d*%s' % (root.cost, root.name))
            
        Recursive(self.root)

def Superblock_IPET(cfg_entrance, model, debug_out, udg_out):
    (vertexes, edges, edge_dict) = Common_Setup(cfg_entrance, model)

    superblock_dict = dict()
    wcet = []
    for e in edges:
        e.contexts = []
    for e in edges:
        if e.superblock != None:
            e.superblock.Add_To_Model(model, wcet, superblock_dict)
    for e in edges:
        if len(e.contexts) != 0:
            model.Add_Term(Name(e), "=", '+'.join(e.contexts))

    model.Add_Term("WCET", "=", '+'.join(wcet))

    def Superblock_Result(key, value):
        if ( superblock_dict.has_key(key) ):
            superblock_dict[ key ].Add_Result(key, value)
        else:
            assert False

    if debug_out != None:
        debug_out.write("Edge -> PC\n")
        for e in edges:
            if e.pc != None:
                debug_out.write("%s -> %x\n" % (Name(e), e.pc))

        debug_out.write("\n\n\n")
        debug_out.write("Superblock data (pre-IPET)\n")
        for e in edges:
            if e.superblock != None:
                e.superblock.Draw(debug_out)
        debug_out.write("\n\n\n")
        debug_out.flush()

    wcet = Common_Solver(model, edge_dict, Superblock_Result)

    if debug_out != None:
        debug_out.write("wcet = %d\n" % wcet)
        debug_out.write("Superblock data (post-IPET)\n")
        for e in edges:
            if e.superblock != None:
                e.superblock.Draw(debug_out)
        
        debug_out.write("Superblock data (post-IPET)\n")

    if udg_out != None:
        Draw_Superblock_Graph(cfg_entrance, udg_out,
                lambda sb_edge: sb_edge.frequency)

    return wcet

def Check_Superblocks(cfg, exe_trace_in, debug_out, udg_out):
    (vertexes, edges) = readcfg.Find_V_E(cfg)
    lookup_edge = dict()

    for e in edges:
        if (e.superblock != None):
            if (e.pc != None):
                lookup_edge[ e.pc ] = e
    
    counters = dict()
    def Inc(key):
        counters[ key ] = counters.get(key, 0) + 1
        if debug_out != None:
            debug_out.write('  inc %s\n' % key)

    if debug_out != None:
        debug_out.write('Tracing program\n')
    superblock = None
    sb_edge = None
    minor = 0
    cost = 0
    for pc in exe_trace_in:
        if superblock != None:
            assert sb_edge != None

            if minor < len(sb_edge.edge.path):
                if debug_out != None:
                    debug_out.write('1 %x %s %u\n' % (
                        pc, sb_edge.name, cost))

                assert sb_edge.edge.path[ minor ] == pc
                minor += 1
            else:
                # Advance to new sb_edge!
                if debug_out != None:
                    debug_out.write('  %x new edge\n' % (pc))

                minor = 0
                hit = 0
                edge_list = sb_edge.Get_Next(True)
                for (s2, c, p) in edge_list:
                    if s2.edge.pc == pc:
                        hit += 1
                        cost += c
                        sb_edge = s2
                        Inc(Name(s2.edge))
                        for s3 in p[ 1: ]:
                            Inc(s3.name)

                assert hit <= 1
                if hit == 0:
                    # Advance to new superblock!
                    superblock = None
                else:
                    if debug_out != None:
                        debug_out.write('2 %x %s %u\n' % (
                            pc, sb_edge.name, cost))

                    assert sb_edge.edge.path[ minor ] == pc
                    minor += 1

        if superblock == None:
            # Load new superblock
            #print hex(pc), 'new superblock'
            if debug_out != None:
                debug_out.write('  %x new superblock\n' % (pc))

            e = lookup_edge[ pc ]
            Inc(Name(e))
            superblock = e.superblock
            edge_list = superblock.root.Get_Next(False)
            assert len(edge_list) == 1
            (sb_edge, c, p) = edge_list.pop()
            assert sb_edge.edge != None
            assert sb_edge.edge == e, (
                "sb_edge.edge.pc = %x  e.pc = %x" % (
                    sb_edge.edge.pc, e.pc))
            cost += c
            minor = 0
            for s3 in p:
                Inc(s3.name)

            if debug_out != None:
                debug_out.write('3 %x %s %u\n' % (
                    pc, sb_edge.name, cost))

            assert sb_edge.edge.path[ minor ] == pc
            minor += 1

    if debug_out != None:
        debug_out.write('Done\n\n')

    if udg_out != None:
        Draw_Superblock_Graph(cfg, udg_out,
                lambda sb_edge: counters.get(sb_edge.name, 0))

    return cost

def Draw_Superblock_Graph(cfg, udg_out, get_frequency_fn):
    (vertexes, edges) = readcfg.Find_V_E(cfg)

    udg_edge_list_dict = dict()
    for e in edges:
        if ((e.superblock != None)
        and (e.pc != None)):
            e.superblock.Collect_Edges(udg_edge_list_dict)

    def Edge_Label(e):
        out = [ 'E%06x' % e.id ]
        if e.pc != None:
            out.append(' at %x size %u' % (e.pc, len(e.path)))

        el = udg_edge_list_dict.get(e.id, [])
        if len(el) == 0:
            out.append('\\nno superblocks')

        for sb_edge in el:
            out.append('\\n%s - %d' % (sb_edge.name, 
                        get_frequency_fn(sb_edge)))

        if e.loop_body != None:
            out.append('\\nloop header')

        if e.branch_type == BT_Serialising:
            out.append('\\nserialising')
        elif e.branch_type == BT_Unconditional:
            out.append('\\nunconditional branch')
        elif e.branch_type == BT_Conditional:
            out.append('\\nconditional branch, %d exits' % len(e.end.exits))

        if (e.frequency != None) and (e.max_frequency != None):
            out.append('\\nfreq %d max %d' % (
                e.frequency, e.max_frequency))

        return ''.join(out)

    for v in vertexes:
        v.Clear_UDG_Flag()

    udg_out.write('[')
    comma = False
    v = vertexes[ 0 ]
    v.Export_UDG(udg_out, Edge_Label)
    udg_out.write(']\n\n')
        
    
